import logging

from abc import ABC, abstractmethod

from threading import Thread, Timer

import pika
from pika import SelectConnection
from pika.exceptions import (
    ConnectionClosedByBroker,
    StreamLostError,
    ConnectionWrongStateError
)

from rabbitmq_client.defs import DEFAULT_EXCHANGE, PublishParams


LOGGER = logging.getLogger(__name__)

RECONNECT_REASONS = (
    ConnectionClosedByBroker,  # Restarts/graceful shutdowns
    StreamLostError  # Ungraceful shutdown
)


class RMQConnection(ABC):
    """
    Abstract base class to be implemented.

    Wraps a pika.SelectConnection and channel object to provide easy access to
    consumer/producer implementers. Allows for a quick start to writing
    consumers and producers, without having to re-write the connection logic.

    RMQConnection signals implementers with the on_ready and on_close methods
    when the connection/channel state changes.
    """

    def __init__(self, connection_parameters=None):
        """
        :param connection_parameters: pika.ConnectionParameters
        """
        # Kept public to allow restart with different parameters.
        self.connection_parameters: pika.ConnectionParameters = \
            connection_parameters

        self._connection_thread = Thread(target=self._connect,
                                         daemon=True)
        self._connection = None
        self._channel = None

        self._reconnect_attempts = 0

        self._closing = False
        self._restarting = False

    @abstractmethod
    def on_ready(self):
        """
        Implementers can handle what should happen when the RMQConnection is
        ready for work.
        """
        pass

    @abstractmethod
    def on_close(self, permanent=False):
        """
        Implementers can handle what should happen when the RMQConnection is
        suddenly closed. An on_close call will normally be followed by an
        on_ready call from the RMQConnection unless the connection is
        permanently down. It is possible that the RMQConnection will signal
        multiple on_close calls for a single connection failure, therefore the
        implementation of on_close must be idempotent.

        :param permanent: bool
        """
        pass

    @abstractmethod
    def on_error(self):
        """
        Implementers can handle what should happen when the RMQConnection has
        run into an error, for example an Exchange/Queue declaration failure.
        """
        pass

    def start(self):
        """
        Starts the connection, opens a connection and a channel by running the
        underlying connection thread. When ready, callbacks will be invoked on
        each step of the way towards establishing the connection and channel,
        starting with on_connection_open if all went well.

        NOTE! Starting the connection is not a synchronous operation! It is
        this way since starting may or may not be successful, so a return
        value would be necessary to indicate success/failure. But, since
        retries will be made, an initial failure may in fact be successful the
        next second, meaning the user is given faulty information. Start
        success is therefore indicated via the 'on_ready' hook instead.
        """
        if not self._connection_thread.is_alive():
            self._connection_thread.start()

    def restart(self):
        """
        Re-initiates the connection and channel.
        """
        LOGGER.info("restarting connection")

        try:
            self._connection.close()
            self._restarting = True  # Only if connection could be closed.
        except ConnectionWrongStateError:
            LOGGER.error("connection closed and could not be restarted")

    def stop(self):
        """
        Starts a closing procedure for the connection and channel.
        """
        LOGGER.info("stopping connection")

        if not self._closing:
            self._closing = True

            try:
                self._connection.close()
                self._connection_thread.join()
            except ConnectionWrongStateError:
                LOGGER.info("connection already closed")

    def declare_queue(self,
                      queue_params,
                      callback=None):
        """
        :param queue_params: rabbitmq_client.QueueParams
        :param callback: callable
        """
        LOGGER.info(f"declaring queue: {queue_params.queue}")

        self._channel.queue_declare(
            queue_params.queue,
            durable=queue_params.durable,
            exclusive=queue_params.exclusive,
            auto_delete=queue_params.auto_delete,
            arguments=queue_params.arguments,
            callback=callback
        )

    def declare_exchange(self,
                         exchange_params,
                         callback=None):
        """
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param callback: callable
        """
        LOGGER.info(f"declaring exchange: {exchange_params.exchange}")

        self._channel.exchange_declare(
            exchange_params.exchange,
            exchange_type=exchange_params.exchange_type,
            durable=exchange_params.durable,
            auto_delete=exchange_params.auto_delete,
            internal=exchange_params.internal,
            arguments=exchange_params.arguments,
            callback=callback
        )

    def bind_queue(self,
                   queue_bind_params,
                   callback=None):
        """
        :param queue_bind_params: rabbitmq_client.QueueBindParams
        :param callback: callable
        """
        LOGGER.info(f"binding queue {queue_bind_params.queue} to exchange "
                    f"{queue_bind_params.exchange} with routing key: "
                    f"{queue_bind_params.routing_key}")

        self._channel.queue_bind(
            queue_bind_params.queue,
            queue_bind_params.exchange,
            routing_key=queue_bind_params.routing_key,
            arguments=queue_bind_params.arguments,
            callback=callback
        )

    def basic_consume(self,
                      consume_params,
                      on_message_callback_override=None,
                      callback=None):
        """
        :param consume_params: rabbitmq_client.ConsumeParams
        :param on_message_callback_override: callable
        :param callback: callable
        """
        LOGGER.info(f"basic consumer starting for queue: "
                    f"{consume_params.queue}")

        self._channel.basic_consume(
            consume_params.queue,
            (on_message_callback_override if on_message_callback_override
             else consume_params.on_message_callback),
            auto_ack=consume_params.auto_ack,
            exclusive=consume_params.exclusive,
            consumer_tag=consume_params.consumer_tag,
            arguments=consume_params.arguments,
            callback=callback
        )

    def basic_publish(self,
                      body,
                      exchange=DEFAULT_EXCHANGE,
                      routing_key="",
                      publish_params=None):
        """
        :param body: bytes
        :param exchange: str
        :param routing_key: str
        :param publish_params: rabbitmq_client.PublishParams
        """
        LOGGER.debug(f"publishing to exchange: {exchange} and routing key: "
                     f"{routing_key}")

        if not publish_params:
            publish_params = PublishParams()  # Create defaults

        self._channel.basic_publish(exchange,
                                    routing_key,
                                    body,
                                    properties=publish_params.properties,
                                    mandatory=publish_params.mandatory)

    def confirm_delivery(self, on_delivery_confirmed, callback=None):
        """
        :param on_delivery_confirmed: callable
        :param callback: callable
        """
        LOGGER.info("activating confirm delivery mode")

        self._channel.confirm_delivery(on_delivery_confirmed,
                                       callback=callback)

    def _connect(self):
        """
        Assigns a new pika.SelectConnection to the connection and starts its
        ioloop to begin connecting.
        """
        if self.connection_parameters is not None:
            LOGGER.info(f"starting connection towards: "
                        f"{self.connection_parameters.host}:"
                        f"{self.connection_parameters.port}")
        else:
            LOGGER.info("starting connection towards: 127.0.0.1:5672")

        self._connection = SelectConnection(
            parameters=self.connection_parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed
        )
        self._connection.ioloop.start()

    def on_connection_open(self, _connection):
        """
        Called when a new connection has been established. Opens a new channel.

        :param _connection: pika.SelectConnection
        """
        LOGGER.debug("connection opened")
        # Do NOT reset reconnect counter here, do it after channel opened.

        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_connection_open_error(self, _connection, error):
        """
        Called when a connection could not be established towards RabbitMQ.
        This will lead to attempts to reconnect with an increasing delay
        between attempts.

        :param _connection: pika.SelectConnection
        :param error: pika.exceptions.?
        """
        LOGGER.warning(f"error establishing connection, error: {error}")

        # This should ensure the current thread runs to completion after
        # reconnect handling is done.
        self._connection.ioloop.stop()

        # No fancy handling, just retry even if credentials are bad.
        self._reconnect()

    def on_connection_closed(self, _connection, reason):
        """
        Called when the connection is closed, either unexpectedly or on
        purpose. Reconnection is attempted if the disconnect reason is
        something that is possible to recover, like an unexpected broker
        shutdown.

        :param _connection: pika.SelectConnection
        :param reason: pika.exceptions.?
        """
        permanent = False

        # This should ensure the current thread runs to completion after
        # on_connection_closed handling is done.
        self._connection.ioloop.stop()

        if not self._closing and type(reason) in RECONNECT_REASONS:
            LOGGER.debug(f"connection closed: {reason}, attempting reconnect")
            self._reconnect()

        elif self._restarting:
            LOGGER.debug("connection closed, restarting now")
            # Reset to prevent reconnect for any reason
            self._restarting = False
            self._connection_thread = Thread(target=self._connect,
                                             daemon=True)
            self._connection_thread.start()

        else:
            permanent = True
            LOGGER.warning(f"connection closed: {reason}, will not reconnect")

        # Signal subclass that connection is down.
        self.on_close(permanent=permanent)

    @property
    def _reconnect_delay(self):
        return self._reconnect_attempts if self._reconnect_attempts < 9 else 30

    def _reconnect(self):
        """
        Starts up the connection again after a gradually increasing delay.
        """
        LOGGER.debug(f"reconnect, attempt no. {self._reconnect_attempts + 1}")

        # Reconnect attempt may have been queued up before 'stop' was called.
        if self._closing:
            LOGGER.debug("skipping reconnect, connection stopped")
            return

        # _connect will assign a new connection object
        self._connection_thread = Thread(target=self._connect,
                                         daemon=True)
        if self._reconnect_delay == 0:  # retry immediately the first time
            self._connection_thread.start()

        else:  # calculate backoff timer
            timer = Timer(
                self._reconnect_delay,
                self._connection_thread.start
            )
            # Timer daemon exits if program exits, to avoid eternal retries.
            timer.daemon = True
            timer.start()

        self._reconnect_attempts += 1

    def on_channel_open(self, channel):
        """
        Called when a new channel has been opened, the consumer is now ready
        for work.

        :param channel: pika.channel.Channel
        """
        LOGGER.debug("channel opened")
        self._reconnect_attempts = 0

        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)

        self.on_ready()  # Signal subclass that connection is ready.

    def on_channel_closed(self, _channel, reason):
        """
        Called when a channel is closed by RabbitMQ. Channels are not
        attempted to be re-opened since they are either closed from going down
        together with the connection, or because input parameters for some
        queue/exchange/other establishment was given faulty parameters.
        Retrying such an operation will just result in a failure loop.

        :param _channel: pika.channel.Channel
        :param reason: pika.exceptions.?
        """
        # TODO: handle 405 - resource locked, also a reason for permanent
        #  closure

        LOGGER.warning(f"channel closed: {reason}")

        permanent = False

        try:
            # Reply code 406 = PRECONDITION FAILED
            permanent = True if reason.reply_code == 406 else False
        except AttributeError:  # Not all reasons have a reply code
            pass

        # Signal subclass that connection is down.
        self.on_close(permanent=permanent)

        if permanent:
            LOGGER.critical(f"connection stopping due to permanent channel "
                            f"closure: {reason}")
            self.stop()



import functools
import logging
import uuid

from threading import Lock

from pika.spec import Basic

from rabbitmq_client.connection import RMQConnection
from rabbitmq_client.defs import (
    ConfirmModeOK,
    DeliveryError,
    DEFAULT_EXCHANGE
)

LOGGER = logging.getLogger(__name__)


class RMQPublish:

    def __init__(self,
                 body,
                 exchange_params=None,
                 routing_key="",
                 queue_params=None,
                 publish_params=None,
                 publish_key=None):
        """
        :param body: bytes
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        :param queue_params: rabbitmq_client.QueueParams
        :param publish_params: rabbitmq_client.PublishParams
        :param publish_key: str
        """
        self.body = body
        self.exchange_params = exchange_params
        self.routing_key = routing_key
        self.queue_params = queue_params
        self.publish_params = publish_params
        self.publish_key = publish_key


class RMQProducer(RMQConnection):
    """
    Generic producer implementation using the RMQConnection base class to ease
    connection and channel handling.
    """

    def __init__(self, connection_parameters=None):
        """
        :param connection_parameters: pika.ConnectionParameters
        """
        super().__init__(connection_parameters=connection_parameters)

        self._ready = False

        self._declared_queues = set()
        self._declared_exchanges = set()

        self._buffered_messages = list()

        # Confirm mode
        self._confirm_mode_active = False
        self._confirm_delivery_callback = None
        self._next_delivery_tag = 1
        self._publish_lock = Lock()
        self._unacked_publishes = dict()

    @property
    def ready(self):
        """
        Indicates if the consumer is ready, meaning it will immediately issue
        consume-work it receives. If the consumer is NOT ready, incoming
        consumes will be delayed until the underlying connection reports ready
        through the 'on_ready' hook.

        :return: bool
        """
        return self._ready

    def start(self):
        """
        Starts the consumer by initiating the underlying connection.

        NOTE! This is NOT a synchronous operation! If you must know that the
        consumer is successfully started, monitor the 'ready' property.
        """
        LOGGER.info("starting producer")

        super().start()

    def restart(self):
        """Restarts the consumer by re-initiating the underlying connection."""
        LOGGER.info("restarting producer")

        super().restart()

    def stop(self):
        """Stops the consumer by stopping the underlying connection."""
        LOGGER.info("stopping producer")

        super().stop()

    def publish(self,
                body,
                exchange_params=None,
                routing_key="",
                queue_params=None,
                publish_params=None):
        """
        Publish 'body' towards an exchange (optional routing key) or a queue.
        You must provide EITHER exchange_params OR queue_params to handle
        the declaration of the publish target, but not both.

        The reasoning here is that a publisher may produce a message to an
        exchange, but is not responsible for declaring the end-queue and its
        binding. So, publish only declares the entity to which it produces its
        message(s) but should not know all the details of an exchange-bound
        queues properties. For work queues, the producer of course needs to
        know declaration details since there is no intermediate step to publish
        to.

        If confirm mode is activated, a key is returned which corresponds to
        the message being published. Users can confirm that their publish has
        been sent successfully by waiting for their key to be referred to in
        a call to their confirm mode notify callback where the input parameter
        == the generated publish key.

        :param body: bytes
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        :param queue_params: rabbitmq_client.QueueParams
        :param publish_params: rabbitmq_client.PublishParams
        :returns: str
        """
        # More detailed logging done on connection level
        LOGGER.info("publishing")

        if (
                (not exchange_params and not queue_params) or
                (exchange_params and queue_params)
        ):
            raise ValueError(
                "You need to provide either a queue OR an exchange, else "
                "there is nothing to publish to..."
            )

        publish_key = None
        if self._confirm_delivery_callback is not None:
            publish_key = str(uuid.uuid4())

        if (
                self.ready and
                (
                    (self._confirm_delivery_callback is not None and
                     self._confirm_mode_active) or
                    self._confirm_delivery_callback is None
                )
        ):
            self._handle_publish(body,
                                 exchange_params,
                                 routing_key,
                                 queue_params,
                                 publish_params,
                                 publish_key)

        else:
            self._buffered_messages.append(
                RMQPublish(body,
                           exchange_params=exchange_params,
                           routing_key=routing_key,
                           queue_params=queue_params,
                           publish_params=publish_params,
                           publish_key=publish_key)
            )

        return publish_key

    def activate_confirm_mode(self, notify_callback):
        """
        :param notify_callback: callable
        """
        if self.ready and self._confirm_delivery_callback is None:
            LOGGER.info("activating confirm delivery mode")
            self.confirm_delivery(self.on_delivery_confirmed,
                                  callback=self.on_confirm_select_ok)

        self._confirm_delivery_callback = notify_callback

    def _handle_publish(self,
                        body,
                        exchange_params,
                        routing_key,
                        queue_params,
                        publish_params,
                        publish_key):
        """
        :param body: bytes
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        :param queue_params: rabbitmq_client.QueueParams
        :param publish_params: rabbitmq_client.PublishParams
        :param publish_key: str
        """
        if (
                queue_params and
                queue_params.queue not in self._declared_queues
        ):

            cb = functools.partial(self.on_queue_declared,
                                   body,
                                   publish_params=publish_params,
                                   publish_key=publish_key)

            self.declare_queue(queue_params, callback=cb)

        elif (
                exchange_params and
                exchange_params.exchange not in self._declared_exchanges
        ):
            cb = functools.partial(self.on_exchange_declared,
                                   body,
                                   exchange_params,
                                   routing_key=routing_key,
                                   publish_params=publish_params,
                                   publish_key=publish_key)

            self.declare_exchange(exchange_params, callback=cb)

        else:  # Declarations are already complete.
            self._finalize_publish(
                body,
                exchange=(DEFAULT_EXCHANGE if exchange_params is None
                          else exchange_params.exchange),
                routing_key=(routing_key if queue_params is None
                             else queue_params.queue),
                publish_params=publish_params,
                publish_key=publish_key
            )

    def on_queue_declared(self,
                          body,
                          frame,
                          publish_params=None,
                          publish_key=None):
        """
        :param body: bytes
        :param frame: pika.frame.Method
        :param publish_params: rabbitmq_client.PublishParams
        :param publish_key: str
        """
        LOGGER.info(f"declared queue: {frame.method.queue}")

        self._declared_queues.add(frame.method.queue)

        self._finalize_publish(body,
                               routing_key=frame.method.queue,
                               publish_params=publish_params,
                               publish_key=publish_key)

    def on_exchange_declared(self,
                             body,
                             exchange_params,
                             _frame,
                             routing_key="",
                             publish_params=None,
                             publish_key=None):
        """
        :param body: bytes
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param _frame: pika.frame.Method
        :param routing_key: str
        :param publish_params: rabbitmq_client.PublishParams
        :param publish_key: str
        """
        LOGGER.info(f"declared exchange: {exchange_params.exchange}")

        self._declared_exchanges.add(exchange_params.exchange)

        self._finalize_publish(body,
                               exchange=exchange_params.exchange,
                               routing_key=routing_key,
                               publish_params=publish_params,
                               publish_key=publish_key)

    def _finalize_publish(self,
                          body,
                          exchange=DEFAULT_EXCHANGE,
                          routing_key="",
                          publish_params=None,
                          publish_key=None):
        """
        :param body: bytes
        :param exchange: str
        :param routing_key: str
        :param publish_params: rabbitmq_client.PublishParams
        :param publish_key: str
        """

        if publish_key is not None:
            with self._publish_lock:
                self._unacked_publishes[self._next_delivery_tag] = publish_key
                self._next_delivery_tag += 1

                self.basic_publish(body,
                                   exchange=exchange,
                                   routing_key=routing_key,
                                   publish_params=publish_params)

        else:
            self.basic_publish(body,
                               exchange=exchange,
                               routing_key=routing_key,
                               publish_params=publish_params)

    def _empty_buffered_messages(self):
        """
        Sends all buffered messages.
        """
        for buffered_message in self._buffered_messages:
            self._handle_publish(buffered_message.body,
                                 buffered_message.exchange_params,
                                 buffered_message.routing_key,
                                 buffered_message.queue_params,
                                 buffered_message.publish_params,
                                 buffered_message.publish_key)

        self._buffered_messages = list()

    def on_confirm_select_ok(self, _frame):
        """
        :param _frame: pika.frame.Method
        """
        LOGGER.info("confirm select ok")
        self._confirm_mode_active = True
        self._confirm_delivery_callback(ConfirmModeOK())

        self._empty_buffered_messages()

    def on_delivery_confirmed(self, frame):
        """
        :param frame: pika.frame.Method
        """
        publish_key = self._unacked_publishes.pop(
            frame.method.delivery_tag
        )

        if isinstance(frame.method, Basic.Ack):
            self._confirm_delivery_callback(publish_key)

        else:
            LOGGER.error(f"broker nacked a publish: {frame}")
            self._confirm_delivery_callback(
                DeliveryError(publish_key)
            )

    def on_ready(self):
        """
        Connection hook, called when channel opened, meaning RMQConnection is
        ready for work.
        """
        LOGGER.info("producer connection ready")

        self._ready = True

        if self._confirm_delivery_callback is not None:
            self.confirm_delivery(self.on_delivery_confirmed,
                                  callback=self.on_confirm_select_ok)

        else:
            self._empty_buffered_messages()

    def on_close(self, permanent=False):
        """
        Connection hook, called when the channel or connection is closed.

        NOTE!
        This is NOT reported as an error because users are expected to
        configure their exchanges and queues accordingly to avoid losing data
        they cannot live without.

        :param permanent: bool
        """
        if not self._closing and permanent:
            LOGGER.critical("producer connection permanently closed")
        else:
            LOGGER.info("producer connection closed")

        self._ready = False
        self._confirm_mode_active = False
        # Delivery tags are channel-specific, so connection going down means
        # the tag is reset.
        self._next_delivery_tag = 1

        # Reset cached declarations, no way of knowing what's still around
        # after a restart.
        self._declared_queues = set()
        self._declared_exchanges = set()

    def on_error(self):
        """
        Connection hook, called when the connection has encountered an error.
        """
        LOGGER.info("producer connection error")

        raise NotImplementedError
        # Possible errors:
        # * channel died and will not recover
        # * callback for operation failed
        # * declaration with faulty parameters attempted
        # TODO: Add possibility to signal user that an error has occurred.



import logging

from rabbitmq_client.connection import RMQConnection  # noqa
from rabbitmq_client.consumer import RMQConsumer  # noqa
from rabbitmq_client.producer import RMQProducer  # noqa
from rabbitmq_client.defs import (  # noqa
    QueueParams,
    ExchangeParams,
    QueueBindParams,
    ConsumeParams,
    PublishParams,
    ConsumeOK,
    ConfirmModeOK,
    DeliveryError,
    DEFAULT_EXCHANGE,
)


logging.getLogger(__name__).addHandler(logging.NullHandler())



from pika.exchange_type import ExchangeType
from typing import Union

import functools
import logging

from rabbitmq_client.defs import (
    QueueParams,
    QueueBindParams,
    ConsumeOK,
    ConsumeParams,
    ExchangeParams
)
from rabbitmq_client.connection import RMQConnection


LOGGER = logging.getLogger(__name__)


def _gen_consume_key(queue: str = "",
                     exchange: str = "",
                     routing_key: str = "") -> str:
    key_list = []
    key_list.append(queue) if queue else ""
    key_list.append(exchange) if exchange else ""
    key_list.append(routing_key) if routing_key else ""

    separator = "|"

    return separator.join(key_list)


class RMQConsume:
    """
    Used by the RMQConsumer to keep track of which consumes have been started
    so far, and to be able to re-distribute consumes should the underlying
    connection experience a hiccup.
    """

    def __init__(self,
                 consume_params,
                 queue_params,
                 exchange_params,
                 routing_key):
        """
        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: rabbitmq_client.QueueParams
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        """
        self.consume_params = consume_params
        self.queue_params = queue_params
        self.exchange_params = exchange_params
        self.routing_key = routing_key

        # The actual consumer tag from 'basic_consume'. This will be equal to
        # 'self.consume_params.consumer_tag' if one was set by the caller,
        # otherwise it is randomly generated by pika. It is used to cancel
        # consumes with 'basic_cancel'.
        self._consumer_tag = None

    @property
    def consumer_tag(self):
        return self._consumer_tag

    @consumer_tag.setter
    def consumer_tag(self, new_value):
        self._consumer_tag = new_value


class RMQConsumer(RMQConnection):
    """
    Generic consumer implementation using the RMQConnection base class to ease
    connection and channel handling.

    The 'consume' method provided by this consumer implementation accepts
    parameters as they will be passed to the pika library to avoid going out
    of its way to create another layer of special handling. A few parameters
    have been removed, but nothing has been added.
    """

    def __init__(self, connection_parameters=None):
        """
        :param connection_parameters: pika.ConnectionParameters
        """
        super().__init__(connection_parameters=connection_parameters)

        self._ready = False
        self._consumes = dict()

        self._declared_queues = set()
        self._declared_exchanges = set()
        self._consumed_queues = dict()

    @property
    def ready(self):
        """
        Indicates if the consumer is ready, meaning it will immediately issue
        consume-work it receives. If the consumer is NOT ready, incoming
        consumes will be delayed until the underlying connection reports ready
        through the 'on_ready' hook.

        :return: bool
        """
        return self._ready

    def start(self):
        """
        Starts the consumer by initiating the underlying connection.

        NOTE! This is NOT a synchronous operation! If you must know that the
        consumer is successfully started, monitor the 'ready' property.
        """
        LOGGER.info("starting consumer")

        super().start()

    def restart(self):
        """Restarts the consumer by re-initiating the underlying connection."""
        LOGGER.info("restarting consumer")

        super().restart()

    def stop(self):
        """Stops the consumer by stopping the underlying connection."""
        LOGGER.info("stopping consumer")

        super().stop()

    def consume(self,
                consume_params,
                queue_params=None,
                exchange_params=None,
                routing_key=None):
        """
        General consumer interface, when wanting to consumer messages sent to a
        specific queue or exchange. Input parameters are a subset of those used
        by pika, to avoid re-inventing the wheel with special abstractions.
        The only parameters that have been removed are the "passive" and
        "callback" options.

        Some restrictions apply:
        1. EITHER queue OR exchange OR both MUST be set.
        2. Queue/exchange parameter change is not supported. For example, this
           consumer will not handle a re-declaration of a queue with new
           provided parameters.

        RMQConsumer will automatically re-issue queue and exchange declarations
        and bindings if the connection is for some reason closed and re-opened
        at any point.

        This method may be called before the underlying RMQConnection has
        successfully established a connection and channel to RabbitMQ. Those
        buffered actions called before RMQConnection reports it is ready,
        through the 'on_ready' hook, will be issued immediately upon 'on_ready'
        being called. Monitor the RMQConsumer 'ready' property to keep track of
        when the RMQConnection has reported it is ready.

        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: rabbitmq_client.QueueParams
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        :returns: str
        """
        LOGGER.info("starting consume")

        # 1. Checks
        if queue_params is None and exchange_params is None:
            raise ValueError(
                "You need to provide either a queue, an exchange, or both, "
                "else there is nothing to consume from..."
            )

        consume_key = _gen_consume_key(
            queue=queue_params.queue if queue_params else "",
            exchange=exchange_params.exchange if exchange_params else "",
            routing_key=routing_key
        )
        if self._consumes.get(consume_key) is not None:
            raise ValueError(
                "That combination of queue + exchange + routing key is "
                "already consumed from."
            )

        if (
                exchange_params is not None and
                exchange_params.exchange_type is ExchangeType.direct and
                not routing_key
        ):
            raise ValueError(
                "Direct exchanges must have a routing key set."
            )

        # 2. Update consumer instance
        self._consumes[consume_key] = RMQConsume(
            consume_params, queue_params, exchange_params, routing_key,
        )

        # 3. Start declaring shit
        if self.ready:
            self._handle_consume(consume_params,
                                 queue_params,
                                 exchange_params,
                                 routing_key)

        return consume_key

    def _handle_consume(self,
                        consume_params,
                        queue_params,
                        exchange_params,
                        routing_key):
        """
        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: None | rabbitmq_client.QueueParams
        :param exchange_params: None | rabbitmq_client.ExchangeParams
        :param routing_key: None | str
        """
        if queue_params is None:
            queue_params = QueueParams("", exclusive=True)

        if queue_params.queue in self._declared_queues:
            consume_params.queue = queue_params.queue

            self.check_declare_exchange(consume_params,
                                        queue_params,
                                        exchange_params,
                                        routing_key)

        else:
            cb = functools.partial(self.on_queue_declared,
                                   consume_params=consume_params,
                                   queue_params=queue_params,
                                   exchange_params=exchange_params,
                                   routing_key=routing_key)

            self.declare_queue(queue_params, callback=cb)

    def on_queue_declared(self,
                          frame,
                          consume_params=None,
                          queue_params=None,
                          exchange_params=None,
                          routing_key=None):
        """
        :param frame: pika.frame.Method
        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: rabbitmq_client.QueueParams
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param routing_key: str
        """
        LOGGER.info(f"declared queue: {frame.method.queue}")

        self._declared_queues.add(frame.method.queue)

        # Update the consume queue name to ensure it is set to the created
        # queue's name in case of automatic name generation.
        consume_params.queue = frame.method.queue

        self.check_declare_exchange(consume_params,
                                    queue_params,
                                    exchange_params,
                                    routing_key)

    def check_declare_exchange(self,
                               consume_params: ConsumeParams,
                               queue_params: QueueParams,
                               exchange_params: Union[ExchangeParams, None],
                               routing_key: str):
        if (
                exchange_params is not None and
                exchange_params.exchange in self._declared_exchanges
        ):
            self.handle_queue_binding(consume_params,
                                      queue_params,
                                      exchange_params,
                                      routing_key)

        elif exchange_params is not None:
            cb = functools.partial(self.on_exchange_declared,
                                   exchange_params,
                                   consume_params=consume_params,
                                   queue_params=queue_params,
                                   routing_key=routing_key)

            self.declare_exchange(exchange_params, callback=cb)

        else:
            ctag = self.ongoing_consume(consume_params.queue)
            if ctag is not None:
                LOGGER.info(f"queue '{consume_params.queue}' "
                            f"is already consumed from")
                # Notify of ongoing consumer tag.
                consume_params.on_message_callback(ConsumeOK(ctag))
                # No need to consume again since the queue is already consumed
                # from.
                return

            cb = functools.partial(self.on_consume_ok,
                                   queue_params=queue_params)

            self.basic_consume(consume_params,
                               on_message_callback_override=self.on_msg,
                               callback=cb)

    def on_exchange_declared(self,
                             exchange_params,
                             _frame,
                             consume_params=None,
                             queue_params=None,
                             routing_key=None):
        """
        :param exchange_params: rabbitmq_client.ExchangeParams
        :param _frame: pika.frame.Method
        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: rabbitmq_client.QueueParams
        :param routing_key: str
        """
        LOGGER.info(f"declared exchange: {exchange_params.exchange}")

        self._declared_exchanges.add(exchange_params.exchange)

        self.handle_queue_binding(consume_params,
                                  queue_params,
                                  exchange_params,
                                  routing_key)

    def handle_queue_binding(self,
                             consume_params: ConsumeParams,
                             queue_params: QueueParams,
                             exchange_params: ExchangeParams,
                             routing_key: str):
        cb = functools.partial(self.on_queue_bound,
                               consume_params=consume_params,
                               queue_params=queue_params,
                               exchange=exchange_params.exchange,
                               routing_key=routing_key)

        self.bind_queue(QueueBindParams(consume_params.queue,
                                        exchange_params.exchange,
                                        routing_key=routing_key),
                        callback=cb)

    def on_queue_bound(self,
                       _frame,
                       consume_params=None,
                       queue_params=None,
                       exchange=None,
                       routing_key=None):
        """
        :param _frame: pika.frame.Method
        :param consume_params: rabbitmq_client.ConsumeParams
        :param queue_params: rabbitmq_client.QueueParams
        :param exchange: str
        :param routing_key: str
        """
        LOGGER.info(f"queue {consume_params.queue} bound to exchange "
                    f"{exchange}")

        ctag = self.ongoing_consume(consume_params.queue)
        if ctag is not None:
            LOGGER.info(f"queue '{consume_params.queue}' "
                        f"is already consumed from")
            # Notify of ongoing consumer tag.
            consume_params.on_message_callback(ConsumeOK(ctag))
            # No need to consume again since the queue is already consumed
            # from.
            return

        cb = functools.partial(self.on_consume_ok,
                               queue_params=queue_params,
                               exchange=exchange,
                               routing_key=routing_key)

        self.basic_consume(consume_params,
                           on_message_callback_override=self.on_msg,
                           callback=cb)

    def ongoing_consume(self, queue: str) -> Union[str, None]:
        return self._consumed_queues.get(queue)

    def on_consume_ok(self,
                      frame,
                      queue_params=None,
                      exchange=None,
                      routing_key=None):
        """
        :param frame: pika.frame.Method
        :param queue_params: rabbitmq_client.QueueParams
        :param exchange: str
        :param routing_key: str
        """
        LOGGER.info(f"consume OK for queue: {queue_params.queue}")

        consume_instance = self._consumes[
            _gen_consume_key(queue=queue_params.queue,
                             exchange=exchange,
                             routing_key=routing_key)
        ]
        # Update with real consumer tag, may or may not be the same tag.
        consume_instance.consumer_tag = frame.method.consumer_tag

        self._consumed_queues[consume_instance.consume_params.queue] = (
            consume_instance.consumer_tag
        )

        # Enables lookup via consumer tag in 'on_msg'. These entries are
        # removed 'on_close' since the consumer tags may be refreshed on
        # reconnecting.
        self._consumes[consume_instance.consumer_tag] = consume_instance

        try:
            consume_instance.consume_params.on_message_callback(
                ConsumeOK(consume_instance.consumer_tag)
            )
        except Exception as e:
            LOGGER.critical(f"sending consume OK to message callback resulted "
                            f"in an exception: {e}")

    def on_msg(self, channel, basic_deliver, _basic_properties, body):
        """
        :param channel: pika.channel.Channel
        :param basic_deliver: pika.spec.Basic.Deliver
        :param _basic_properties: pika.spec.Basic.Properties
        :param body: bytes
        """
        LOGGER.debug(f"consumer received message: {body}")

        consume = self._consumes[basic_deliver.consumer_tag]

        try:
            if not consume.consume_params.auto_ack:
                consume.consume_params.on_message_callback(
                    body,
                    ack=lambda:
                        channel.basic_ack(
                            delivery_tag=basic_deliver.delivery_tag
                        )
                )
            else:
                consume.consume_params.on_message_callback(body)
        except Exception as e:
            LOGGER.warning(f"the on_message_callback for queue: "
                           f"{consume.consume_params.queue} "
                           f"crashed with error: {e}")

    def on_ready(self):
        """
        Connection hook, called when channel opened, meaning RMQConnection is
        ready for work.
        """
        LOGGER.info("consumer connection ready")

        self._ready = True

        for _key, consume in self._consumes.items():
            self._handle_consume(consume.consume_params,
                                 consume.queue_params,
                                 consume.exchange_params,
                                 consume.routing_key)

    def on_close(self, permanent=False):
        """
        Connection hook, called when the channel or connection is closed.

        NOTE!
        This is NOT reported as an error because users are expected to
        configure their exchanges and queues accordingly to avoid losing data
        they cannot live without.

        :param permanent: bool
        """
        if not self._closing and permanent:
            LOGGER.critical("consumer connection permanently closed")
        else:
            LOGGER.info("consumer connection closed")

        self._ready = False

        # Reset cached declarations, no way of knowing what's still around
        # after a restart.
        self._declared_queues = set()
        self._declared_exchanges = set()
        self._consumed_queues = dict()

        old_consumer_tags = list()
        for _key, consume in self._consumes.items():
            # Remove all consumer tag keys in local dict since these may change
            # on re-subscription.
            if consume.consumer_tag is not None:
                old_consumer_tags.append(consume.consumer_tag)
            consume.consumer_tag = None

        for consumer_tag in old_consumer_tags:
            self._consumes.pop(consumer_tag, None)

    def on_error(self):
        """
        Connection hook, called when the connection has encountered an error.
        """
        LOGGER.info("consumer connection error")

        raise NotImplementedError
        # Possible errors:
        # * channel died and will not recover
        # * callback for operation failed
        # * declaration with faulty parameters attempted
        # TODO: Add possibility to signal user that an error has occurred.



from pika.exchange_type import ExchangeType


#
# DEFS
#
DEFAULT_EXCHANGE = ""


#
# STRUCTS
#
class QueueParams:

    def __init__(self,
                 queue,
                 durable=False,
                 exclusive=False,
                 auto_delete=False,
                 arguments=None):
        """
        # noqa 501
        https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.queue_declare

        Excluding the "passive" and "callback" options. Passive may not be used
        in this context, and callback is used by the connection object.
        """
        self.queue = queue
        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.arguments = arguments


class ExchangeParams:

    def __init__(self,
                 exchange,
                 exchange_type=ExchangeType.direct,
                 durable=False,
                 auto_delete=False,
                 internal=False,
                 arguments=None):
        """
        # noqa 501
        https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.exchange_declare

        Excluding the "passive" and "callback" options. Passive may not be used
        in this context, and callback is used by the connection object.
        """
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.durable = durable
        self.auto_delete = auto_delete
        self.internal = internal
        self.arguments = arguments


class ConsumeParams:

    def __init__(self,
                 on_message_callback,
                 auto_ack=False,
                 exclusive=False,
                 consumer_tag=None,
                 arguments=None):
        """
        # noqa 501
        https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.basic_consume

        Excluding the "passive" and "callback" options. Passive may not be used
        in this context, and callback is used by the connection object.

        The queue property is for internal use ONLY, and hence is not exposed
        in this class' constructor.

        Consumer tag may be provided to force the use of a particular tag,
        the tag used in the end is communicated to 'on_message_callback' when
        consuming has successfully started through a ConsumeOK class instance.
        """
        self.on_message_callback = on_message_callback
        self.auto_ack = auto_ack
        self.exclusive = exclusive
        self._consumer_tag = consumer_tag
        self.arguments = arguments

        self._queue = None

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, new_value):
        self._queue = new_value

    @property
    def consumer_tag(self):
        return self._consumer_tag

    @consumer_tag.setter
    def consumer_tag(self, new_value):
        self._consumer_tag = new_value


class PublishParams:

    def __init__(self,
                 properties=None,
                 mandatory=False):
        """
        # noqa 501
        https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.basic_publish

        Since body, exchange, and routing key are mandatory parameters of
        basic_publish, they are not included in this parameter object but
        rather a part of the general publish interface of RMQProducer.
        """
        self.properties = properties
        self.mandatory = mandatory


class QueueBindParams:

    def __init__(self,
                 queue,
                 exchange,
                 routing_key=None,
                 arguments=None):
        """
        # noqa 501
        https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.queue_bind
        """
        self.queue = queue
        self.exchange = exchange
        self.routing_key = routing_key
        self.arguments = arguments


class ConsumeOK:

    def __init__(self, consumer_tag):
        """
        :param consumer_tag: str
        """
        self.consumer_tag = consumer_tag

    def __str__(self):
        return f"<ConsumeOK consumer_tag: '{self.consumer_tag}'>"


class ConfirmModeOK:
    pass


class DeliveryError:

    def __init__(self, publish_key):
        """
        :param publish_key: int
        """
        self.publish_key = publish_key
